--[[
Special thanks to:

  * Bartender3 - Several portions of the Bartender3 source code were studied 
  	as a teaching aid for the development of this add-on.  In particular,
  	much of the layout and positioning code was based on Bartender3, with 
  	minor modifications to suit Numen's needs.  Many thanks and praises go 
  	to the Bartender3 Development Team for their work.
  	
  * AutoBar - The original concept of Numen was based around AutoBar.  While 
  	none of the code from AutoBar was used or consulted, without AutoBar, 
  	Numen may never have been created.
  	
  * Klappa - In combination with the WoW Forums, Klappa was used as a teaching 
  	aid for secure frames and state headers.  Many hours of understanding 
  	these components were spared because of Klappa.

  * WoW Forums - Many of the examples of secure frames and state headers found 
  	in the WoW Forums were used as a guide, in conjunction with Klappa, to 
  	create Numen.
  	
  * Ace2 and Libraries - This goes without saying.  Numen would not be here 
  	without the work of the Ace Development Team.  Many thanks goes out to 
  	them for creating a simple and easy framework to build add-ons on.
  
--]]
local _, class = UnitClass("player");
if (class ~= "SHAMAN") then 
	return;
end

Numen = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceDB-2.0", "AceDebug-2.0", "AceEvent-2.0", "FuBarPlugin-2.0")

local L = AceLibrary("AceLocale-2.2"):new("Numen")
local Tablet = AceLibrary("Tablet-2.0")
local Gratuity = AceLibrary("Gratuity-2.0");

local _G = getfenv(0)

--[[
local totems = {
	earth = {
		name = "Earth",
		index = 1,
		totems = {
			tremor = {
				spell = "Tremor Totem",
				index = 1,
			},
			claw = {
				spell = "Stoneclaw Totem",
				index = 2,
			},
			bind = {
				spell = "Earthbind Totem",
				index = 3,
			},
			strength = {
				spell = "Strength of Earth Totem",
				index = 4,
			},
			skin = {
				spell = "Stoneskin Totem",
				index = 5,
			},
			earthelem = {
				spell = "Earth Elemental Totem",
				index = 6,
			},
		},
	},
	fire = {
		name = "Fire",
		index = 2,
		totems = {
			wrath = {
				spell = "Totem of Wrath",
				index = 1,
			},
			sear = {
				spell = "Searing Totem",
				index = 2,
			},
			magma = {
				spell = "Magma Totem",
				index = 3,
			},
			nova = {
				spell = "Fire Nova Totem",
				index = 4,
			},
			flame = {
				spell = "Flametongue Totem",
				index = 5,
			},
			frostres = {
				spell = "Frost Resistance Totem",
				index = 6,
			},
			fireelem = {
				spell = "Fire Elemental Totem",
				index = 7,
			},
		},
	},
	water = {
		name = "Water",
		index = 3,
		totems = {
			manaspr = {
				spell = "Mana Spring Totem",
				index = 1,
			},
			heal = {
				spell = "Healing Stream Totem",
				index = 2,
			},
			poison = {
				spell = "Poison Cleansing Totem",
				index = 3,
			},
			disease = {
				spell = "Disease Cleansing Totem",
				index = 4,
			},
			manatide = {
				spell = "Mana Tide Totem",
				index = 5,
			},
			fireres = {
				spell = "Fire Resistance Totem",
				index = 6,
			},
		},
	},
	air = {
		name = "Air",
		index = 4,
		totems = {
			wrathair = {
				spell = "Wrath of Air Totem",
				index = 1,
			},
			ground = {
				spell = "Grounding Totem",
				index = 2,
			},
			grace = {
				spell = "Grace of Air Totem",
				index = 3,
			},
			windfury = {
				spell = "Windfury Totem",
				index = 4,
			},
			windwall = {
				spell = "Windwall Totem",
				index = 5,
			},
			natureres = {
				spell = "Nature Resistance Totem",
				index = 6,
			},
			tranquil = {
				spell = "Tranquil Air Totem",
				index = 7,
			},
			sentry = {
				spell = "Sentry Totem",
				index = 8,
			},
		},
	},
	weapon = {
		name = "Weapon",
		index = 5,
		totems = {
			flametongue = {
				spell = "Flametongue Weapon",
				index = 1,
			},
			windfury = {
				spell = "Windfury Weapon",
				index = 2,
			},
			rockbiter = {
				spell = "Rockbiter Weapon",
				index = 3,
			},
			frostbrand = {
				spell = "Frostbrand Weapon",
				index = 4,
			},
		},
	},
}
--]]
Numen.totems = {};
Numen.isLocked = true;

Numen.options = {
	type = "group",
	args = {
		bar = {
			type = "group",
			name = "Bar Configuration",
			desc = "Configuration for the Numen Bar.",
			args = {
				vertical = {
					type = "toggle",
					name = "Vertical",
					desc = "Set the orientation of the Numen Bar as vertical.",
					get = function() return Numen.db.profile.isVert; end,
					set = "ToggleVert",
					map = { [false] = "Disabled", [true] = "Enabled" },
				},
				grow = {
					type = "toggle",
					name = "growth",
					desc = "Popup growth",
					get = function() return Numen.db.profile.isRtDw; end,
					set = "ToggleRtDw",
					map = { [false] = "Disabled", [true] = "Enabled" },
				},
			},
		},
	},
}

local defaults = {
	isVert = false,
	isRtDw = false,
	totemStates = {},
}

 -- {{{ FuBar settings
Numen.name                   = "Numen"
Numen.hasIcon                = "Interface\\Icons\\Spell_Fire_TotemOfWrath"
Numen.defaultMinimapPosition = 240
Numen.cannotDetachTooltip    = true
Numen.independentProfile     = true
Numen.defaultPosition        = "LEFT"
Numen.hideWithoutStandby     = true

 -- }}}

function Numen:OnInitialize()
    -- Called when the addon is loaded
	self:RegisterDB("NumenDB", "NumenDBPC")
	self:RegisterDefaults("profile", defaults)
	self:RegisterChatCommand("/numen", self.options)

	self.keyframe = CreateFrame("Frame", nil, UIParent);
	self:RegisterEvent("UPDATE_BINDINGS")
	BINDING_HEADER_Numen = "Numen";
	
	Numen.config = self.db.profile;
	self.OnMenuRequest = self.options;
end

function Numen:OnEnable()
	self:SetDebugging(true)
	self:Debug("Loaded");
	self.totems = self:GetTotemTable();
	self:RegisterOverrideBindings()
	self:InitBar();
	self:LoadPosition();
	self:UpdateLayout();
end

function Numen:UPDATE_BINDINGS()
	if (not InCombatLockdown()) then
		self:RegisterOverrideBindings();
	end
end

function Numen:RegisterOverrideBindings()
	ClearOverrideBindings(self.keyframe)
	for totemType, args in pairs(self.totems) do
		local key1, key2 = GetBindingKey("Numen"..args.name.."Button");
		if key1 then
			SetOverrideBindingClick(self.keyframe, false, key1, "Numen"..args.name.."Button0");
		end
		if key2 then
			SetOverrideBindingClick(self.keyframe, false, key2, "Numen"..args.name.."Button0");
		end
	end
end

function Numen:OnDisable()
    -- Called when the addon is disabled
end

function Numen:OnTooltipUpdate()
	Tablet:SetHint("|cffeda55fLeft Click|r to lock Numen bar.\n|cffeda55fRight-Click|r to open config.")
end

function Numen:OnClick(button)
	-- function for fubar left click
	self:ToggleLock();
	self:UpdateDisplay();
end

function Numen:ToggleVert(value)
	self.db.profile.isVert = value;
	self:UpdateLayout();
	if (value) then
		self.options.args.bar.args.grow.name = "Grow Right";
		self.options.args.bar.args.grow.desc = "Set the popup buttons to grow to the right.";
	else
		self.options.args.bar.args.grow.name = "Grow Down";
		self.options.args.bar.args.grow.desc = "Set the popup buttons to grow downward.";
	end
end

function Numen:ToggleRtDw(value)
	self.db.profile.isRtDw = value;
	self:UpdateLayout();
end

function Numen:UpdateLayout()
	local rootx, rooty = 0, 0;
	local isVert = self.db.profile.isVert;
	local isRtDw = self.db.profile.isRtDw;
	if (isVert) then
		rootx = 36;
	else
		rooty = 36;
	end
	for tfidx, totemFrame in pairs(self.totemFrames) do
		local x, y = 0, 0;
		if (isVert) then
			y = -36*(tfidx);
			rooty = rooty + 36;
		else
			x = 36*(tfidx);
			rootx = rootx + 36;
		end
		totemFrame:SetPoint("TOPLEFT", self.root, "TOPLEFT", x, y);
		for tpidx, totemPop in pairs(self.totemPops[tfidx]) do
			local x, y = 0, 0;
			if (isVert) then
				x = 36*tpidx;
				if (not isRtDw) then
					x = x*(-1);
				end
			else
				y = -36*tpidx;
				if (not isRtDw) then
					y = y*(-1);
				end
			end
			totemPop:SetPoint("TOPLEFT", totemFrame, "TOPLEFT", x, y);
		end
	end
	self.root:SetWidth(rootx);
	self.root:SetHeight(rooty);
end

function Numen:InitBar()
	self.root = CreateFrame("Frame", "NumenRoot", UIParent, "SecureStateHeaderTemplate");
	self.totemFrames = {};
	self.totemPops = {};
	
	for totemType, typeArgs in pairs(self.totems) do
		local tfidx = typeArgs.index-1;
		self.totemPops[tfidx] = {};
		self.totemFrames[tfidx] = CreateFrame("Frame", "Numen"..typeArgs.name.."Frame", self.root, "SecureStateHeaderTemplate");
		totemButton = CreateFrame("CheckButton", "Numen"..typeArgs.name.."Button0", self.totemFrames[tfidx], "SecureAnchorEnterTemplate,SecureActionButtonTemplate,ActionButtonTemplate");
		totemButton:SetPoint("TOPLEFT", self.totemFrames[tfidx], "TOPLEFT", 0, 0);
		totemButton:SetWidth(36);
		totemButton:SetHeight(36);
		totemButton.cooldown = _G["Numen"..typeArgs.name.."Button0Cooldown"];
		totemButton:SetAttribute("type", "spell");
		totemButton:SetAttribute("childstate-onenter", "show"..totemType);
		totemButton:SetAttribute("childstate-onleave", "hide"..totemType);
		totemButton:SetAttribute("anchorchild", self.root);
		totemButton:SetAttribute("totemType", typeArgs.name);
		totemButton:SetAttribute("tfidx", tfidx);
		totemButton:SetScript("OnAttributeChanged", SwapTexture);
		totemButton:HookScript("OnEnter", OnEnterButton);
		totemButton:HookScript("OnLeave", OnLeaveButton);
		totemButton.texture = totemButton:CreateTexture(nil, "BACKGROUND");
		totemButton:SetCheckedTexture("");
		totemButton:SetNormalTexture("");
		DLAdded(totemButton);
		totemButton.texture:SetPoint("TOPLEFT", totemButton, "TOPLEFT", 3, -3);
		totemButton.texture:SetPoint("BOTTOMRIGHT", totemButton, "BOTTOMRIGHT",  -3, 3);
		totemButton.texture:SetTexCoord(0.08, 0.92, 0.08, 0.92);
		self.totemFrames[tfidx]:SetAttribute("addchild", totemButton);
		statestr = "";
		for totem, totemArgs in pairs(typeArgs.totems) do
			local tpidx = totemArgs.index;
			self.totemPops[tfidx][tpidx] = CreateFrame("CheckButton", "Numen"..typeArgs.name.."Button"..totemArgs.index, totemButton, "SecureAnchorUpDownTemplate,SecureActionButtonTemplate,ActionButtonTemplate");
			self.totemPops[tfidx][tpidx]:SetWidth(36);
			self.totemPops[tfidx][tpidx]:SetHeight(36);
			self.totemPops[tfidx][tpidx].texture = self.totemPops[tfidx][tpidx]:CreateTexture(nil, "BACKGROUND");
			self.totemPops[tfidx][tpidx].texture:SetAllPoints(self.totemPops[tfidx][tpidx]);
			self.totemPops[tfidx][tpidx].texture:SetTexture(GetSpellTexture(totem));
			self.totemPops[tfidx][tpidx]:SetCheckedTexture("");
			self.totemPops[tfidx][tpidx]:SetNormalTexture("");
			DLAdded(self.totemPops[tfidx][tpidx]);
			self.totemPops[tfidx][tpidx].texture:SetPoint("TOPLEFT", self.totemPops[tfidx][tpidx], "TOPLEFT", 3, -3);
			self.totemPops[tfidx][tpidx].texture:SetPoint("BOTTOMRIGHT", self.totemPops[tfidx][tpidx], "BOTTOMRIGHT",  -3, 3);
			self.totemPops[tfidx][tpidx].texture:SetTexCoord(0.08, 0.92, 0.08, 0.92);
			self.totemPops[tfidx][tpidx]:SetAttribute("*type1", "spell");
			self.totemPops[tfidx][tpidx]:SetAttribute("spell", totem);
			self.totemPops[tfidx][tpidx].Spell = totem;
			self.totemPops[tfidx][tpidx]:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN", "UpdateCooldown");
			self.totemPops[tfidx][tpidx]:SetAttribute("onmouseupbutton", "mup");
			self.totemPops[tfidx][tpidx]:SetAttribute("*childstate-mup", totemType..totemArgs.index);
			self.totemPops[tfidx][tpidx]:SetAttribute("shift-childstate-mup", ATTRIBUTE_NOOP);
			self.totemPops[tfidx][tpidx]:SetAttribute("anchorchild", self.totemFrames[tfidx]);
			self.totemPops[tfidx][tpidx]:SetAttribute("showstates", tostring(typeArgs.index));
			totemButton:SetAttribute("spell-"..totemType..totemArgs.index, totem);
			self.totemFrames[tfidx]:SetAttribute("statemap-anchor-"..totemType..totemArgs.index, tostring(totemArgs.index));
			self.root:SetAttribute("addchild", self.totemPops[tfidx][tpidx]);
			statestr = statestr..totemArgs.index..":"..totemType..totemArgs.index..";";
			self.totemPops[tfidx][tpidx]:SetAttribute("spellid", totemArgs.spellid);
			self.totemPops[tfidx][tpidx]:SetScript("OnEnter", OnEnterButton);
			self.totemPops[tfidx][tpidx]:SetScript("OnLeave", OnLeaveButton);
		end
		totemButton:SetAttribute("statebutton", statestr);
		self.totemFrames[tfidx]:SetAttribute("state", self.db.profile.totemStates[tfidx] or "1");
		self.root:SetAttribute("statemap-anchor-show"..totemType, "*:"..typeArgs.index);
		self.root:SetAttribute("statemap-anchor-hide"..totemType, ";");
		self.root:SetAttribute("delaystatemap-anchor-hide"..totemType, typeArgs.index..":0");
		self.root:SetAttribute("delaytimemap-anchor-hide"..totemType,  typeArgs.index..":0.5");
		self.root:SetAttribute("delayhovermap-anchor-hide"..totemType, typeArgs.index..":true");
	end
	self.root:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
	self.root:RegisterForDrag("LeftButton");
	self.root:EnableMouse(false);
	self.root:SetMovable(true);
	self.root.texture = self.root:CreateTexture();
	self.root.texture:SetTexture(0, 0, 0.5, 0);
	self.root.texture:SetAllPoints(self.root);

end

function DLAdded(button)
	local frame = CreateFrame("Frame", button:GetName().."DL", button)
	frame:ClearAllPoints()
	frame:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 1, edgeFile = "", edgeSize = 0, insets = {left = 0, right = 0, top = 0, bottom = 0},})
	frame:SetBackdropColor(0, 0, 0, 0.6)
	frame:SetAllPoints(button)
	frame:SetFrameLevel(0)
	return frame

end

function SwapTexture(button, name, value)
	if (name=="state-parent") then
		if (value ~= "0") then 
			totemType = button:GetAttribute("totemType");
			tfidx = button:GetAttribute("tfidx");
--			Numen:Debug(tfidx..":"..tonumber(value));
			if (Numen.totemPops[tfidx][tonumber(value)]) then
				popButton = Numen.totemPops[tfidx][tonumber(value)];
				button.texture:SetTexture(popButton.texture:GetTexture());
				button:SetAttribute("spellid", popButton:GetAttribute("spellid"));
			end
			Numen.db.profile.totemStates[tfidx] = value;
		end
	end
end

function OnEnterButton(button)
	if (button:GetAttribute("spellid")) then
		--GameTooltip:SetOwner(button, "ANCHOR_RIGHT");
		GameTooltip:SetOwner(button);
		GameTooltip:SetSpell(button:GetAttribute("spellid"), BOOKTYPE_SPELL);
	end
end

function OnLeaveButton(button)
	GameTooltip:Hide();
end

function Numen:ToggleLock()
	if (self.isLocked) then
		self.root:EnableMouse(true);
		self.root:SetScript("OnDragStart", StartDrag);
		self.root:SetScript("OnDragStop", StopDrag);
		self.root.texture:SetTexture(0, 0, 0.5, 0.5)
		self.root:SetFrameLevel(4);
		self.isLocked = false;
	else
		self.root:EnableMouse(false);
		self.root:SetScript("OnDragStart", nil);
		self.root:SetScript("OnDragStop", nil);
		self.root.texture:SetTexture(0, 0, 0.5, 0);
		self.isLocked = true;
	end		
end

function StartDrag()
	Numen.root:StartMoving();
end

function StopDrag()
	Numen.root:StopMovingOrSizing();
	Numen:SavePosition();
	
end

function Numen:LoadPosition()
	local x, y, s = self.config.rootX, self.config.rootY, self.root:GetEffectiveScale();
	if x and y then 
		x, y = x/s, y/s;
	end
	self.root:ClearAllPoints();
	self.root:SetPoint(x and "TOPLEFT" or "CENTER", UIParent, x and "BOTTOMLEFT" or "CENTER", x or 0, y or 0);
end

function Numen:SavePosition()
	local x, y, s = self.root:GetLeft(), self.root:GetTop(), self.root:GetEffectiveScale();
	x, y = x*s, y*s;
	self.config.rootX = x;
	self.config.rootY = y;
end
------------------------------------------
--  Debugging Functions			--
------------------------------------------

function UpdateCooldown()
	Numen:Debug(elsapsed);
end

function Numen:TestIndy(indicator, type, number)
	if (not indicator)  then
		indicator = "corner1";
		type = 0;
		number = 2;
	end
	testButton = CreateFrame("Frame", "bobby", self.totemPops[type][number]);
	testButton:SetWidth(8);
	testButton:SetHeight(8);
	testButton:SetBackdrop( {
		      bgFile = "Interface\\Addons\\Grid\\white16x16", tile = true, tileSize = 16,
		      edgeFile = "Interface\\Addons\\Grid\\white16x16", edgeSize = 1,
		      insets = {left = 1, right = 1, top = 1, bottom = 1},
	      });
	testButton:SetBackdropBorderColor(0,0,0,1);
	testButton:SetBackdropColor(1,1,1,1);
	testButton:SetFrameLevel(5);
	if indicator == "corner4" then
		testButton:SetPoint("BOTTOMLEFT", self.totemPops[type][number], "BOTTOMLEFT", 1, 1)
	elseif indicator == "corner3" then
		testButton:SetPoint("BOTTOMRIGHT", self.totemPops[type][number], "BOTTOMRIGHT", -1, 1)
	elseif indicator == "corner2" then
		testButton:SetPoint("TOPRIGHT", self.totemPops[type][number], "TOPRIGHT", -1, -1)
	elseif indicator == "corner1" then
		testButton:SetPoint("TOPLEFT", self.totemPops[type][number], "TOPLEFT", 1, -1)
	end
	testButton:Show();
end
function Numen:MoveIndy(indicator, type, number)
	testButton = bobby;
	testButton:ClearAllPoints();
	if indicator == "corner4" then
		testButton:SetPoint("BOTTOMLEFT", self.totemPops[type][number], "BOTTOMLEFT", 1, 1)
	elseif indicator == "corner3" then
		testButton:SetPoint("BOTTOMRIGHT", self.totemPops[type][number], "BOTTOMRIGHT", -1, 1)
	elseif indicator == "corner2" then
		testButton:SetPoint("TOPRIGHT", self.totemPops[type][number], "TOPRIGHT", -1, -1)
	elseif indicator == "corner1" then
		testButton:SetPoint("TOPLEFT", self.totemPops[type][number], "TOPLEFT", 1, -1)
	end
end

function RecAddonMsg(arg1, arg2, arg3, arg4)
  if (arg1 == "NumenMsg") then
    if (string.find(arg2, "broadcast")) then
      --  ******player sending is arg4
      --  ******get earth index
        _, _, earth = string.find(arg2,"e:(.)")
      --  ******get fire index
        _, _, fire = string.find(arg2,"f:(.)")
      --  ******get water index
        _, _, water = string.find(arg2,"w:(.)")
      --  ******get air index
        _, _, air = string.find(arg2,"a:(.)")
    elseif (string.find(arg2, "change")) then
        _, _, typeChanged, index = string.find(arg2,"(.):(.)")
      --  ******typeChanged = e, f, w, or a
    end
  end
  
  --[[ event coding: 
		EVENT: "PARTY_MEMBERS_CHANGED" {
		--set all players in group using numen to 0
		SendAddonMessage("NumenMsg", "broadcast, e:1; f:1; w:1; a:1", "PARTY");
		}

		EVENT: numen click buttons {
		SendAddonMessage("NumenMsg", "change, e:2", "PARTY");
		}
  --]]
end

function Numen:GetTotemTable()
	totemTable = {
		earth = {
			name = "Earth",
			index = 1,
			count = 0,
			totems = {
			},
		},
		fire = {
			name = "Fire",
			index = 2,
			count = 0,
			totems = {
			},
		},
		water = {
			name = "Water",
			index = 3,
			count = 0,
			totems = {
			},
		},
		air = {
			name = "Air",
			index = 4,
			count = 0,
			totems = {
			},
		},
		weapon = {
			name = "Weapon",
			index = 5,
			count = 0,
			totems = {
			},
		},
	};
	spellID = 1;
	repeat
		Gratuity:SetSpell(spellID, BOOKTYPE_SPELL);
		if (Gratuity:Find("Tools: ", 4, 4, false, true) or Gratuity:Find(" Weapon", 1, 1, false, true)) then
			spellName, spellRank = GetSpellName(spellID, BOOKTYPE_SPELL);
			toolLine = Gratuity:GetLine(4);
			local totemType;
			if (string.find(toolLine, "Earth Totem")) then
				totemType = "earth";
			elseif (string.find(toolLine, "Fire Totem")) then
				totemType = "fire";
			elseif (string.find(toolLine, "Water Totem")) then
				totemType = "water";
			elseif (string.find(toolLine, "Air Totem")) then
				totemType = "air";
			elseif (string.find(spellName, " Weapon")) then
				totemType = "weapon";
			end
			if (totemType) then
				spellRank = tonumber(strsub(spellRank, 5)) or 0;
				if (totemTable[totemType].totems[spellName]) then
					if (spellRank > totemTable[totemType].totems[spellName].rank) then
						totemTable[totemType].totems[spellName].rank = spellRank;
						totemTable[totemType].totems[spellName].spellid = spellID;
					end
				else
					totemIdx = totemTable[totemType].count + 1;
					totemTable[totemType].count = totemIdx;
					totemTable[totemType].totems[spellName] = {
						rank = spellRank;
						spellid = spellID;
						index = totemIdx;
					}
				end
			end
		end
		spellID = spellID + 1;
	until (Gratuity:NumLines() == 0);
	return totemTable;
end
